package ru.dbotthepony.mc.otm.block.entity.matter

import net.minecraft.core.BlockPos
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.block.matter.MatterBottlerBlock
import ru.dbotthepony.mc.otm.block.entity.MatteryPoweredBlockEntity
import ru.dbotthepony.mc.otm.block.entity.WorkerState
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.item.ProxiedItemHandler
import ru.dbotthepony.mc.otm.capability.UpgradeType
import ru.dbotthepony.mc.otm.capability.energy.ProfiledEnergyStorage
import ru.dbotthepony.mc.otm.capability.energy.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.capability.matter.MatterStorageImpl
import ru.dbotthepony.mc.otm.capability.matter.ProfiledMatterStorage
import ru.dbotthepony.mc.otm.config.MachinesConfig
import ru.dbotthepony.mc.otm.container.UpgradeContainer
import ru.dbotthepony.mc.otm.container.slotted.FilteredContainerSlot
import ru.dbotthepony.mc.otm.container.slotted.SlottedContainer
import ru.dbotthepony.mc.otm.menu.matter.MatterBottlerMenu
import ru.dbotthepony.mc.otm.registry.game.MBlockEntities
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.util.countingLazy
import ru.dbotthepony.mc.otm.graph.matter.SimpleMatterNode
import java.util.function.BooleanSupplier

class MatterBottlerBlockEntity(blockPos: BlockPos, blockState: BlockState) :
	MatteryPoweredBlockEntity(MBlockEntities.MATTER_BOTTLER, blockPos, blockState) {

	val upgrades = UpgradeContainer(3, UpgradeType.BASIC_MATTER, BooleanSupplier { this.blockState.getValue(WorkerState.SEMI_WORKER_STATE) !== WorkerState.IDLE }, ::markDirtyFast)
	override val energy = ProfiledEnergyStorage(WorkerEnergyStorage(this::markDirtyFast, upgrades.transform(MachinesConfig.MatterBottler.VALUES)))
	val energyConfig = ConfigurableEnergy(energy)

	private inner class BottlingSlot(container: SlottedContainer, slot: Int) : FilteredContainerSlot(container, slot) {
		override fun canAutomationPlaceItem(itemStack: ItemStack): Boolean {
			return super.canAutomationPlaceItem(itemStack) && isBottling && itemStack.getCapability(MatteryCapability.MATTER_ITEM)?.receiveMatterChecked(Decimal.ONE, true)?.isPositive == true
		}
	}

	private inner class UnBottlingSlot(container: SlottedContainer, slot: Int) : FilteredContainerSlot(container, slot) {
		override fun canAutomationPlaceItem(itemStack: ItemStack): Boolean {
			return super.canAutomationPlaceItem(itemStack) && !isBottling && itemStack.getCapability(MatteryCapability.MATTER_ITEM)?.extractMatterChecked(Decimal.ONE, true)?.isPositive == true
		}
	}

	val unbottling = SlottedContainer.simple(3, ::UnBottlingSlot, ::updateBlockState)
	val bottling = SlottedContainer.simple(3, ::BottlingSlot, ::updateBlockState)

	var isBottling: Boolean = true
		set(value) {
			field = value
			initialCapacity = null
			workProgress = 0f
			this.markDirtyFast()

			if (value) {
				inputHandler.parent = bottling
				outputHandler.parent = unbottling
			} else {
				inputHandler.parent = unbottling
				outputHandler.parent = bottling
			}

			updateBlockState()
		}

	var spitItemsWhenCantWork = false
		set(value) {
			field = value
			this.markDirtyFast()
		}

	private val inputHandler = ProxiedItemHandler(bottling)
	private val outputHandler = ProxiedItemHandler(unbottling)

	val itemConfig = ConfigurableItemHandler(
		input = inputHandler,
		output = outputHandler,
		battery = batteryContainer
	)

	val matter: ProfiledMatterStorage<MatterStorageImpl> = ProfiledMatterStorage(object : MatterStorageImpl(this@MatterBottlerBlockEntity::markDirtyFast, FlowDirection.BI_DIRECTIONAL, upgrades.matterCapacity(MachinesConfig.MatterBottler.VALUES::matterCapacity)) {
		override val matterFlow: FlowDirection get() {
			return if (this@MatterBottlerBlockEntity.isBottling) FlowDirection.INPUT else FlowDirection.OUTPUT
		}
	})

	val matterNode = SimpleMatterNode(matter = matter)

	init {
		exposeGlobally(MatteryCapability.MATTER_BLOCK, matter)
		exposeGlobally(MatteryCapability.MATTER_NODE, matterNode)

		savetablesConfig.bool(::isBottling)
		savetablesConfig.bool(::spitItemsWhenCantWork)
		savetables.stateful(::energy, ENERGY_KEY)
		savetables.stateful(::matter, MATTER_STORAGE_KEY)
		savetables.stateful(::bottling)
		savetables.stateful(::unbottling)
		savetables.stateful(::upgrades)
	}

	var workProgress: Float = 0f
		private set

	private var lastTickWorking = false
	private var initialCapacity: Decimal? = null

	override fun setLevel(level: Level) {
		super.setLevel(level)
		matterNode.discover(this)
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MatterBottlerMenu(containerID, inventory, this)
	}

	override fun setRemoved() {
		super.setRemoved()
		matterNode.isValid = false
	}

	private fun updateBlockState(container: SlottedContainer) {
		val level = level as? ServerLevel ?: return
		var state = blockState

		for ((i, slot) in container.slotIterator().withIndex()) {
			val desired = !slot.isEmpty && slot.item.getCapability(MatteryCapability.MATTER_ITEM) != null

			if (state.getValue(MatterBottlerBlock.SLOT_PROPERTIES[i]) != desired) {
				state = state.setValue(MatterBottlerBlock.SLOT_PROPERTIES[i], desired)
			}
		}

		if (state !== blockState) {
			level.setBlock(blockPos, state, Block.UPDATE_ALL)
		}
	}

	private fun updateBlockState() {
		markDirtyFast()
		if (isBottling)
			updateBlockState(bottling)
		else
			updateBlockState(unbottling)
	}

	private val workerState by countingLazy(blockStateChangesCounter) {
		this.blockState.getValue(WorkerState.SEMI_WORKER_STATE)
	}

	private fun blockstateToWorking() {
		if (workerState !== WorkerState.WORKING) {
			level?.setBlock(blockPos, blockState.setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.WORKING), Block.UPDATE_ALL)
		}
	}

	private fun blockstateToIdle() {
		if (workerState !== WorkerState.IDLE) {
			level?.setBlock(blockPos, blockState.setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.IDLE), Block.UPDATE_ALL)
		}
	}

	private fun tickBottling() {
		var isWorking = false
		var hasCapacitors = false

		for (slot in bottling.slotIterator()) {
			val item = slot.item
			val capability = item.getCapability(MatteryCapability.MATTER_ITEM) ?: continue

			if (!capability.receiveMatterChecked(Decimal.LONG_MAX_VALUE, true).isPositive) {
				unbottling.consumeItem(item, false)
				slot.setChanged()
				continue
			}

			hasCapacitors = true
			initialCapacity = initialCapacity ?: capability.storedMatter
			val rate = MachinesConfig.MatterBottler.RATE * (1.0 + upgrades.speedBonus)
			val energyRate = MachinesConfig.MatterBottler.VALUES.energyConsumption * (1.0 + upgrades.speedBonus)

			if (matter.storedMatter < rate) {
				val toExtract = matter.missingMatter
					.coerceAtMost(rate * 200)
					.coerceAtMost(capability.missingMatter - matter.storedMatter)

				matter.receiveMatter(matterNode.graph.extractMatter(toExtract, false), false)
			}

			if (matter.storedMatter.isPositive) {
				val energyRatio = if (energyRate <= Decimal.ZERO) Decimal.ONE else energy.extractEnergy(energyRate, true) / energyRate
				val matterRatio = matter.extractMatter(capability.receiveMatter(rate.coerceAtMost(matter.storedMatter), true), true) / rate

				val minRatio = minOf(matterRatio, energyRatio)

				if (minRatio > Decimal.ZERO) {
					isWorking = true
					energy.extractEnergy(energyRate * minRatio, false)
					matter.extractMatter(capability.receiveMatter(rate * minRatio, false), false)
					workProgress = ((capability.storedMatter - initialCapacity!!) / capability.maxStoredMatter).toFloat()
					slot.setChanged()
				}
			} else {
				if (spitItemsWhenCantWork) {
					unbottling.consumeItem(item, false)
					slot.setChanged()
				}
			}

			break
		}

		if (isWorking) {
			blockstateToWorking()
		} else {
			blockstateToIdle()
		}

		if (!hasCapacitors) {
			matter.extractMatter(matterNode.graph.receiveMatter(matter.storedMatter, false), false)
			initialCapacity = null
			workProgress = 0f
		}
	}

	private fun tickUnbottling() {
		matter.extractMatter(matterNode.graph.receiveMatter(matter.storedMatter, false), false)

		if (!matter.missingMatter.isPositive) {
			if (spitItemsWhenCantWork) {
				for (slot in unbottling.slotIterator()) {
					bottling.consumeItem(slot.item, false)
					slot.setChanged()
				}
			}

			blockstateToIdle()
			return
		}

		var any = false
		var hasCapacitors = false

		for (slot in unbottling.slotIterator()) {
			val item = slot.item
			val it = item.getCapability(MatteryCapability.MATTER_ITEM) ?: continue

			if (!it.extractMatterChecked(Decimal.LONG_MAX_VALUE, true).isPositive) {
				bottling.consumeItem(item, false)
				slot.setChanged()
				continue
			}

			initialCapacity = initialCapacity ?: it.storedMatter

			hasCapacitors = true
			val rate = MachinesConfig.MatterBottler.RATE * (1.0 + upgrades.speedBonus)
			val energyRate = MachinesConfig.MatterBottler.VALUES.energyConsumption * (1.0 + upgrades.speedBonus)

			val energyRatio = if (energyRate <= Decimal.ZERO) Decimal.ONE else energy.extractEnergy(energyRate, true) / energyRate
			val matterRatio = matter.receiveMatter(it.extractMatterChecked(rate, true), true) / rate

			val minRatio = minOf(energyRatio, matterRatio)

			if (minRatio > Decimal.ZERO) {
				any = true
				energy.extractEnergy(energyRate * energyRatio, false)
				matter.receiveMatter(it.extractMatterChecked(rate * minRatio, false), false)

				workProgress = 1f - (it.storedMatter / initialCapacity!!).toFloat()
			}

			break
		}

		if (any) {
			blockstateToWorking()
		} else {
			blockstateToIdle()
		}

		if (!hasCapacitors) {
			initialCapacity = null
			workProgress = 0f
		}
	}

	override fun tick() {
		super.tick()

		if (redstoneControl.isBlockedByRedstone) {
			blockstateToIdle()
			return
		}

		if (isBottling) {
			tickBottling()
		} else {
			tickUnbottling()
		}
	}
}
